open Printf ;;

module type TREE = sig
  type t
  type label
  val children : t -> t list
  val label : t -> label
  val string_of_label : label -> string
end ;;

module type EPS_TREE = sig
  type t
  type label
  val children : t -> t list
  val label : t -> label
  val string_of_label : label -> string

  type color =
    | Gray of float
    | RGB of float * float * float

  type eps_info = {
      rlimit : float ;
      width : float ;
      height : float ;
      foreground : color ;
      background : color ;
      edge_width : float ;
      edge_color : color ;
      label_foreground : color ;
      label_background : color ;
      font_name : string ;
      font_size : float
    }

  type coord = float * float

  val write : string -> eps_info -> t -> coord -> float -> unit
end ;;

module Make(T : TREE) = struct
  type t = T.t
  type label = T.label
  let children = T.children
  let label = T.label
  let string_of_label = T.string_of_label

  module HT = Htree.Make(T)

  type color =
    | Gray of float
    | RGB of float * float * float

  let color_string = function
    | Gray f -> sprintf "%g setgray" f
    | RGB(r, g, b) -> sprintf "%g %g %g setrgbcolor" r g b

  type eps_info = {
      rlimit : float ;
      width : float ;
      height : float ;
      foreground : color ;
      background : color ;
      edge_width : float ;
      edge_color : color ;
      label_foreground : color ;
      label_background : color ;
      font_name : string ;
      font_size : float
    }

  type coord = float * float

  let write filename info t start angle =
    let ch = open_out filename in
    let xzoom = info.width/.2.0
    and yzoom = info.height/.2.0 in
    let convert (zx, zy) =
      (zx*.xzoom +. xzoom, zy*.yzoom +. yzoom) in
    let moveto z =
      let (x, y) = convert z in
      fprintf ch "%g %g m\n" x y
    and lineto z =
      let (x, y) = convert z in
      fprintf ch "%g %g l\n" x y
    and curveto z1 z2 z3 =
      let (x1, y1) = convert z1
      and (x2, y2) = convert z2
      and (x3, y3) = convert z3 in
      fprintf ch "%g %g %g %g %g %g c\n" x1 y1 x2 y2 x3 y3
    and draw_label lab z f =
      let (x, y) = convert z in
      fprintf ch "%g %g (%s) t\n" x y (string_of_label lab) in
    let drv =
      { HT.rlimit = info.rlimit ;
	HT.moveto = moveto ;
	HT.lineto = lineto ;
	HT.curveto = curveto ;
	HT.draw_label = draw_label ;
	HT.init_edge_pass = ignore ;
	HT.init_label_pass = ignore ;
	HT.finalize = ignore } in
    (*** Writing the header ***)
    fprintf ch "%%!PS-Adobe-2.0 EPSF-2.0\n" ;
    fprintf ch "%%%%Creator: EpsTree\n" ;
    fprintf ch "%%%%BoundingBox: 0 0 %g %g\n" info.width info.height ;
    fprintf ch "%%%%DocumentFonts: %s\n" info.font_name ;
    fprintf ch "%%%%EndComments\n" ;
    (*** Writing the preamble ***)
    fprintf ch "/m { moveto } def\n" ;
    fprintf ch "/l { lineto stroke } def\n" ;
    fprintf ch "/c { curveto stroke } def\n" ;
    fprintf ch "/t {\n" ;
    fprintf ch "  /text exch def /y exch def /x exch def\n" ;
    fprintf ch "  text stringwidth pop\n" ;
    fprintf ch "  /width exch def /height %g def\n" info.font_size ;
    fprintf ch "  %s\n" (color_string info.label_background) ;
    fprintf ch "  x width 2 div sub 2 sub\n" ;
    fprintf ch "  y height 2 div sub 2 sub moveto\n" ;
    fprintf ch "  width 4 add 0 rlineto\n" ;
    fprintf ch "  0 height 4 add rlineto\n" ;
    fprintf ch "  width 4 add neg 0 rlineto\n" ;
    fprintf ch "  closepath fill\n" ;
    fprintf ch "  %s\n" (color_string info.label_foreground) ;
    fprintf ch "  x width 2 div sub\n" ;
    fprintf ch "  y height 2 div sub moveto\n" ;
    fprintf ch "  text show\n" ;
    fprintf ch "} def\n" ;
    fprintf ch "/%s findfont %g scalefont setfont\n"
      info.font_name info.font_size ;
    fprintf ch "%s\n" (color_string info.edge_color) ;
    fprintf ch "%g setlinewidth\n" info.edge_width ;
    (*** Writing the body ***)
    HT.draw_curved_tree drv t start angle ;
    close_out ch
end ;;
