open Graphics ;;

(* Uncomment the following two lines if you only have
   a pre-3.0 version of the Objective Caml compiler *)

(*
let synchronize _ = () ;;
let auto_synchronize _ = () ;;
*)

module type TREE = sig
  type t
  type label
  val children : t -> t list
  val label : t -> label
  val string_of_label : label -> string
end

module type GTREE = sig
  type t
  type label
  val children : t -> t list
  val label : t -> label
  val string_of_label : label -> string

  val show_tree : t -> int -> int -> unit
end

let curveto =
  let rec bezier_rec x0 y0 x1 y1 x2 y2 x3 y3 =
    let dx01 = x1 -. x0
    and dy01 = y1 -. y0
    and dx23 = x3 -. x2
    and dy23 = y3 -. y2
    and dx03 = x3 -. x0
    and dy03 = y3 -. y0 in
    let a1 = abs_float (dx01*.dy03 -. dy01*.dx03)
    and a2 = abs_float (dx23*.dy03 -. dy23*.dx03)
    and d = sqrt(dx03*.dx03 +. dy03*.dy03) in
    let amax = max 1.0 (d/.2.0) in
    if a1 < amax && a2 < amax then
      lineto (truncate x3) (truncate y3)
    else
      begin
	let x01 = (x0 +. x1)/.2.0
	and y01 = (y0 +. y1)/.2.0
	and x12 = (x1 +. x2)/.2.0
	and y12 = (y1 +. y2)/.2.0
	and x23 = (x2 +. x3)/.2.0
	and y23 = (y2 +. y3)/.2.0 in
	let x012 = (x01 +. x12)/.2.0
	and y012 = (y01 +. y12)/.2.0
	and x123 = (x12 +. x23)/.2.0
	and y123 = (y12 +. y23)/.2.0 in
	let x0123 = (x012 +. x123)/.2.0
	and y0123 = (y012 +. y123)/.2.0 in
	bezier_rec x0 y0 x01 y01 x012 y012 x0123 y0123 ;
	bezier_rec x0123 y0123 x123 y123 x23 y23 x3 y3
      end in
  fun x1 y1 x2 y2 x3 y3 ->
    let (x0, y0) = current_point () in
    bezier_rec
      (float x0) (float y0)
      (float x1) (float y1)
      (float x2) (float y2)
      (float x3) (float y3) ;;

module Make(T : TREE) = struct
  type t = T.t
  type label = T.label
  let children = T.children
  let label = T.label
  let string_of_label = T.string_of_label

  module HT = Htree.Make(T)
  module ET = EpsTree.Make(T)

  let show_tree t width height =
    let rlimit = 0.98
    and xzoom = float(width)/.2.0
    and yzoom = float(height)/.2.0 in

    let eps_info = {
      ET.rlimit = rlimit ;
      ET.width = float width ;
      ET.height = float height ;
      ET.background = ET.Gray 1.0 ;
      ET.foreground = ET.Gray 0.0 ;
      ET.edge_width = 0.4 ;
      ET.edge_color = ET.Gray 0.0 ;
      ET.label_foreground = ET.Gray 1.0 ;
      ET.label_background = ET.RGB(0.0, 0.0, 1.0) ;
      ET.font_name = "Helvetica-Bold" ;
      ET.font_size = 12.0
    } in

    let common_moveto (zx, zy) =
      let x = truncate(zx*.xzoom +. xzoom)
      and y = truncate(zy*.yzoom +. yzoom) in
      moveto x y

    and common_lineto (zx, zy) =
      let x = truncate(zx*.xzoom +. xzoom)
      and y = truncate(zy*.yzoom +. yzoom) in
      lineto x y in

    let common_curveto (zx1, zy1) (zx2, zy2) (zx3, zy3) =
      let x1 = truncate (zx1*.xzoom +. xzoom)
      and x2 = truncate (zx2*.xzoom +. xzoom)
      and x3 = truncate (zx3*.xzoom +. xzoom)
      and y1 = truncate (zy1*.yzoom +. yzoom)
      and y2 = truncate (zy2*.yzoom +. yzoom)
      and y3 = truncate (zy3*.yzoom +. yzoom) in
      curveto x1 y1 x2 y2 x3 y3 in

    let draw_label lab (zx, zy) _ =
      let x = truncate(zx*.xzoom +. xzoom)
      and y = truncate(zy*.yzoom +. yzoom) in
      let name = string_of_label lab in
      let (w,h) = text_size name in
      let x0 = x - w/2
      and y0 = y - h/2 in
      set_color blue ;
      fill_rect (x0 - 2) (y0 - 1) (w + 4) (h + 2) ;
      moveto x0 y0 ;
      set_color white ;
      draw_string name

    and draw_init_edge_pass () = set_color black

    and draw_init_label_pass () = () in

    let draw_drv = {
      HT.rlimit = rlimit ;
      HT.moveto = common_moveto ;
      HT.lineto = common_lineto ;
      HT.curveto = common_curveto ;
      HT.draw_label = draw_label ;
      HT.init_edge_pass = draw_init_edge_pass ;
      HT.init_label_pass = draw_init_label_pass ;
      HT.finalize = (fun () -> ())
    } in

    let clear_label lab (zx, zy) _ =
      let x = truncate(zx*.xzoom +. xzoom)
      and y = truncate(zy*.yzoom +. yzoom) in
      let name = string_of_label lab in
      let (w,h) = text_size name in
      let x0 = x - w/2
      and y0 = y - h/2 in
      fill_rect (x0 - 2) (y0 - 1) (w + 4) (h + 2)

    and clear_init_edge_pass () = set_color white

    and clear_init_label_pass () = () in

    let clear_drv = {
      HT.rlimit = rlimit ;
      HT.moveto = common_moveto ;
      HT.lineto = common_lineto ;
      HT.curveto = common_curveto ;
      HT.draw_label = clear_label ;
      HT.init_edge_pass = clear_init_edge_pass ;
      HT.init_label_pass = clear_init_label_pass ;
      HT.finalize = (fun () -> ())
    } in

    let xy2c x y =
      let zx = (float(x) -. xzoom)/.xzoom
      and zy = (float(y) -. yzoom)/.yzoom in
      let zn = sqrt(zx*.zx +. zy*.zy) in
      if zn > rlimit then
	(rlimit*.zx/.zn, rlimit*.zy/.zn)
      else
	(zx, zy)

    and draw_linear_tree = HT.draw_linear_tree draw_drv
    and draw_curved_tree = HT.draw_curved_tree draw_drv
    and clear_linear_tree = HT.draw_linear_tree clear_drv
    and clear_curved_tree = HT.draw_curved_tree clear_drv

    and flags = [Button_down; Button_up; Key_pressed; Mouse_motion]
    and origin = ref (0.0, 0.0)
    and old_xy = ref None
    and continue = ref true in

    let display =
      try Sys.getenv "DISPLAY"
      with Not_found -> ":0.0" in

    open_graph (display ^ " " ^
		string_of_int width ^ "x" ^
		string_of_int height) ;
    auto_synchronize false ;

    let fontname = "-*-helvetica-bold-r-normal-*-12-*-*-*-*-*-*-*" in
    set_font fontname ;
    draw_curved_tree t !origin 0.0 ;
    synchronize () ;

    while !continue do
      let st = wait_next_event flags in
      if st.button then
	begin
          match !old_xy with
          | None ->
              clear_curved_tree t !origin 0.0 ;
              draw_linear_tree t !origin 0.0 ;
	      synchronize () ;
              old_xy := Some(st.mouse_x, st.mouse_y)
          | Some(x,y) ->
              let z1 = xy2c x y
              and z2 = xy2c st.mouse_x st.mouse_y in
              clear_linear_tree t !origin 0.0 ;
              origin := HT.drag_origin !origin z1 z2 ;
              draw_linear_tree t !origin 0.0 ;
	      synchronize () ;
              old_xy := Some(st.mouse_x, st.mouse_y)
	end
      else
	begin
          match !old_xy with
          | None -> ()
          | Some(x,y) ->
              clear_linear_tree t !origin 0.0 ;
              draw_curved_tree t !origin 0.0 ;
	      synchronize () ;
              old_xy := None
	end ;
      if st.keypressed then
	begin
	  match st.key with
	  | 'c' ->
	      if !old_xy = None then
		begin
		  clear_curved_tree t !origin 0.0 ;
		  origin := (0.0, 0.0) ;
		  draw_curved_tree t !origin 0.0 ;
		  synchronize ()
		end
	  | 's' ->
	      ET.write "htree.eps" eps_info t !origin 0.0
	  | 'q' ->
	      continue := false
	  | _ -> ()
	end
    done ;
    close_graph ()
end
